#drinc:libraries/dos.g
#drinc:exec/miscellaneous.g
#drinc:exec/memory.g
#drinc:graphics/gfx.g
#iff.g
#iffILBM.g

uint
    MAX_LEN = 128,
    WIDTH = 752,
    HEIGHT = 480,
    DEPTH = 24;

uint
    COLOR_TABLE_SIZE = 32,
    COLOR_CYCLE_SIZE = 4;

type
    FatBitMap_t = struct {
	uint fbm_BytesPerRow;
	uint fbm_Rows;
	ushort fbm_Flags;
	ushort fbm_Depth;
	uint fbm_pad;
	[MAX_PLANES]PLANEPTR fbm_Planes;
    },

    ILBMPrivate_t = struct {
	BitMapHeader_t ip_bmh;
	[COLOR_TABLE_SIZE] uint ip_colorMap;
	Point2D_t ip_p2d;
	DestMerge_t ip_dm;
	SpritePrecedence_t ip_sp;
	ViewPortMode_t ip_vpm;
	[COLOR_CYCLE_SIZE] CRange_t ip_cycle;
	ushort ip_colorCount, ip_cycleCount;
    };
	
Handle_t StdOut, OutputFd;
*char OutputName;
[IFF_BUFFER_SIZE] byte Buffer;
uint Position;
[MAX_LEN] byte Dump;
bool DiskFull, LocalError;

proc writeString(*char str)void:
    *char p;

    p := str;
    while p* ~= '\e' do
	p := p + 1;
    od;
    ignore Write(StdOut, str, p - str);
corp;

proc writeCount(uint count)void:
    [6] char buffer;
    *char p;

    p := &buffer[5];
    p* := '\e';
    while
	p := p - 1;
	p* := count % 10 + '0';
	count := count / 10;
	count ~= 0
    do
    od;
    writeString(p);
corp;

proc putByte(byte b)void:

    if not DiskFull then
	if Position = IFF_BUFFER_SIZE then
	    if Write(OutputFd, &Buffer[0], IFF_BUFFER_SIZE) ~=
		    IFF_BUFFER_SIZE then
		DiskFull := true;
	    fi;
	    Position := 0;
	fi;
	Buffer[Position] := b;
	Position := Position + 1;
    fi;
corp;

proc putImage(*FatBitMap_t fbm; *ILBMPrivate_t ip)void:
    register **byte pPlane;
    register uint c, i;
    register byte b, mask;
    uint r;
    [8] *byte planes;

    DeleteFile(OutputName);
    OutputFd := Open(OutputName, MODE_NEWFILE);
    if OutputFd ~= 0 then
	Position := 0;
	for i from 0 upto 7 do
	    planes[i] := pretend(fbm*.fbm_Planes[i], *byte);
	od;
	for r from HEIGHT - 1 downto 0 do
	    mask := 0x80;
	    for c from WIDTH - 1 downto 0 do
		b := 0;
		pPlane := &planes[0];
		for i from 8 - 1 downto 0 do
		    b := b >> 1;
		    if pPlane** & mask ~= 0 then
			b := b | 0x80;
		    fi;
		    pPlane := pPlane + sizeof(*byte);
		od;
		putByte(b);
		mask := mask >> 1;
		if mask = 0 then
		    mask := 0x80;
		    pPlane := &planes[0];
		    for i from 8 - 1 downto 0 do
			pPlane* := pPlane* + 1;
			pPlane := pPlane + sizeof(*byte);
		    od;
		fi;
	    od;
	od;
	if Position ~= 0 and not DiskFull then
	    if Write(OutputFd, &Buffer[0], Position) ~= Position then
		DiskFull := true;
	    fi;
	fi;
	Close(OutputFd);
    fi;
corp;

proc err(*char message; uint was, desired)void:

    if not LocalError then
	LocalError := true;
	writeString(":\n");
    fi;
    writeString(message);
    writeString(" - was ");
    writeCount(was);
    writeString(", must be ");
    writeCount(desired);
    writeString("\n");
corp;

proc readCat(*IFFContext_t context; IFFID_t subId)bool:

    IFFReadCatTail(context)
corp;

proc readList(*IFFContext_t context; IFFID_t subId)bool:
    ILBMPrivate_t ip;
    bool res;

    ip := pretend(context*.iff_client*.iff_userStructure, *ILBMPrivate_t)*;
    res := IFFReadListTail(context);
    pretend(context*.iff_client*.iff_userStructure, *ILBMPrivate_t)* := ip;
    res
corp;

proc readForm(*IFFContext_t context; IFFID_t subId)bool:
    ILBMPrivate_t ip;
    bool res;

    ip := pretend(context*.iff_client*.iff_userStructure, *ILBMPrivate_t)*;
    res := IFFReadFormTail(context);
    pretend(context*.iff_client*.iff_userStructure, *ILBMPrivate_t)* := ip;
    res
corp;

proc readProp(*IFFContext_t context; IFFID_t subId)bool:

    IFFReadPropTail(context)
corp;

proc readOther(*IFFContext_t context; IFFID_t id)bool:
    FatBitMap_t fbm;
    *ILBMPrivate_t ip;
    uint depth, w, h, i;
    bool ok;

    ip := pretend(context*.iff_client*.iff_userStructure, *ILBMPrivate_t);
    case id.iff_ulong
    incase ID_BBMHD:
	if IFFReadBBMHD(context, &ip*.ip_bmh) then
	    if ip*.ip_bmh.bmh_nPlanes > DEPTH then
		err("too many bit planes", ip*.ip_bmh.bmh_nPlanes, DEPTH);
	    elif ip*.ip_bmh.bmh_nPlanes < DEPTH then
		err("too few bit planes", ip*.ip_bmh.bmh_nPlanes, DEPTH);
	    fi;
	    if ip*.ip_bmh.bmh_w > WIDTH then
		err("image too wide", ip*.ip_bmh.bmh_w, WIDTH);
	    elif ip*.ip_bmh.bmh_w < WIDTH then
		err("image too narrow", ip*.ip_bmh.bmh_w, WIDTH);
	    fi;
	    if ip*.ip_bmh.bmh_h > HEIGHT then
		err("image too high", ip*.ip_bmh.bmh_h, HEIGHT);
	    elif ip*.ip_bmh.bmh_h < HEIGHT then
		err("image too low ", ip*.ip_bmh.bmh_h, HEIGHT);
	    fi;
	    not LocalError
	else
	    false
	fi
    incase ID_BCMAP:
	ip*.ip_colorCount := COLOR_TABLE_SIZE;
	IFFReadBCMAP(context, &ip*.ip_colorMap[0], &ip*.ip_colorCount)
    incase ID_BGRAB:
	IFFReadBGRAB(context, &ip*.ip_p2d)
    incase ID_BDEST:
	IFFReadBDEST(context, &ip*.ip_dm)
    incase ID_BSPRT:
	IFFReadBSPRT(context, &ip*.ip_sp)
    incase ID_BCAMG:
	IFFReadBCAMG(context, &ip*.ip_vpm)
    incase ID_BCRNG:
	i := ip*.ip_cycleCount;
	if i = COLOR_CYCLE_SIZE then
	    IFFSkipChunk(context)
	else
	    ok := IFFReadBCRNG(context, &ip*.ip_cycle[i]);
	    if ok and ip*.ip_cycle[i].cr_active ~= 0 and
		    ip*.ip_cycle[i].cr_low ~= ip*.ip_cycle[i].cr_high then
		ip*.ip_cycleCount := i + 1;
	    fi;
	    ok
	fi
    incase ID_BBODY:
	depth := ip*.ip_bmh.bmh_nPlanes;
	w := ip*.ip_bmh.bmh_w;
	h := ip*.ip_bmh.bmh_h;
	InitBitMap(pretend(&fbm, *BitMap_t), depth, w, h);
	ok := true;
	for i from 0 upto depth - 1 do
	    fbm.fbm_Planes[i] := AllocRaster(w, h);
	    if fbm.fbm_Planes[i] = nil then
		ok := false;
	    fi;
	od;
	if ok then
	    ok := IFFReadBBODY(context, pretend(&fbm, *BitMap_t),
			       nil, &ip*.ip_bmh);
	    if ok then
		putImage(&fbm, ip);
	    fi;
	fi;
	for i from depth - 1 downto 0 do
	    if fbm.fbm_Planes[i] ~= nil then
		FreeRaster(fbm.fbm_Planes[i], w, h);
	    fi;
	od;
	ok
    default:
	IFFSkipChunk(context)
    esac
corp;

proc printFailure(IFFError_t errorCode)void:

    writeString(" - failed: ");
    writeString(
	case errorCode
	incase iff_ok:
	    "no error"
	incase iff_noFile:
	    "can't open file"
	incase iff_notIFF:
	    "not IFF file"
	incase iff_notEnough:
	    "file or chunk too short"
	incase iff_tooMuch:
	    "file or chunk too long"
	incase iff_badIFF:
	    "misformed IFF file"
	incase iff_illegalIFF:
	    "illegal structure for IFF file"
	default:
	    "unknown error"
	esac);
    writeString("\n");
corp;

proc usage()void:

    writeString("Use is: Craig iff-file raw-file\n");
corp;

proc main()void:
    extern
	_d_pars_initialize()void,
	GetPar()*char;
    ILBMPrivate_t ip;
    IFFClient_t client;
    uint i;
    *char inputName;

    if OpenDosLibrary(0) ~= nil then
	if OpenExecLibrary(0) ~= nil then
	    if OpenGraphicsLibrary(0) ~= nil then
		StdOut := Output();
		_d_pars_initialize();
		client.iff_readForm := readForm;
		client.iff_readList := readList;
		client.iff_readCat := readCat;
		client.iff_readProp := readProp;
		client.iff_readOther := readOther;
		pretend(client.iff_userStructure, *ILBMPrivate_t) := &ip;
		inputName := GetPar();
		if inputName ~= nil then
		    OutputName := GetPar();
		    if OutputName ~= nil then
			writeString(inputName);
			ip.ip_bmh.bmh_w := 320;
			ip.ip_bmh.bmh_h := 200;
			ip.ip_bmh.bmh_nPlanes := 5;
			ip.ip_bmh.bmh_masking := bmsk_none;
			ip.ip_bmh.bmh_compression := bcmp_byteRun1;
			ip.ip_colorCount := COLOR_TABLE_SIZE;
			ip.ip_cycleCount := 0;
			for i from 0 upto COLOR_CYCLE_SIZE - 1 do
			    ip.ip_cycle[i].cr_active := 0;
			od;
			DiskFull := false;
			LocalError := false;
			if IFFReadFile(inputName, &client) then
			    writeString(" => ");
			    writeString(OutputName);
			    writeString("\n");
			elif DiskFull then
			    writeString(" => ");
			    writeString(OutputName);
			    writeString(": error on write\n");
			elif not LocalError then
			    printFailure(client.iff_error);
			fi;
		    else
			usage();
		    fi;
		else
		    usage();
		fi;
		CloseGraphicsLibrary();
	    fi;
	    CloseExecLibrary();
	fi;
	CloseDosLibrary();
    fi;
corp;
