#drinc:libraries/dos.g
#drinc:exec/miscellaneous.g
#drinc:exec/memory.g
#drinc:graphics/gfx.g
#iff.g
#iffILBM.g

uint DEPTH = 5;

type ILBMPrivate_t = struct {
    BitMapHeader_t ip_bmh;
};
	
*char InputName;
Handle_t StdOut;
[IFF_BUFFER_SIZE] byte Buffer;
bool LocalError;
uint ByteNumber;

proc writeString(*char str)void:
    *char p;

    p := str;
    while p* ~= '\e' do
	p := p + 1;
    od;
    ignore Write(StdOut, str, p - str);
corp;

proc writeCount(uint count)void:
    [6] char buffer;
    *char p;

    p := &buffer[5];
    p* := '\e';
    while
	p := p - 1;
	p* := count % 10 + '0';
	count := count / 10;
	count ~= 0
    do
    od;
    writeString(p);
corp;

proc putPixel(byte colour)void:
    [3] char buffer;
    register byte digit;

    if ByteNumber % 4 = 0 then
	writeString("    tile[");
	writeCount(ByteNumber / 4);
	writeString("] := 0x");
    fi;
    digit := colour >> 4;
    buffer[0] := if digit > 9 then digit - 10 + 'a' else digit + '0' fi;
    digit := colour & 0x0f;
    buffer[1] := if digit > 9 then digit - 10 + 'a' else digit + '0' fi;
    buffer[2] := '\e';
    writeString(&buffer[0]);
    ByteNumber := ByteNumber + 1;
    if ByteNumber % 4 = 0 then
	writeString(";\n");
    fi;
corp;

proc putImage(register *BitMap_t bm; register *ILBMPrivate_t ip)void:
    register *uint p;
    register byte colour;
    register uint shift, offset, k, j, i;

    writeString("define t_tiles proc make");
    if InputName* >= 'a' and InputName* <= 'z' then
	InputName* := InputName* - 'a' + 'A';
    fi;
    writeString(InputName);
    writeString("Tile()list int:\n    list int tile;\n\n");
    writeString("    tile := CreateIntArray(");
    writeCount((ip*.ip_bmh.bmh_h * ip*.ip_bmh.bmh_w + 3) / 4);
    writeString(");\n");
    ByteNumber := 0;
    offset := 0;
    shift := 15;
    for i from 0 upto ip*.ip_bmh.bmh_h - 1 do
	for j from 0 upto ip*.ip_bmh.bmh_w - 1 do
	    colour := 0x00;
	    for k from DEPTH - 1 downto 0 do
		colour := colour << 1;
		colour := colour |
		    ((bm*.bm_Planes[k] + offset)* >> shift) & 0b1;
	    od;
	    putPixel(colour);
	    if shift = 0 then
		shift := 15;
		offset := offset + 2;
	    else
		shift := shift - 1;
	    fi;
	od;
	if shift ~= 15 then
	    shift := 15;
	    offset := offset + 2;
	fi;
    od;
    while ByteNumber % 4 ~= 0 do
	writeString("00");
	ByteNumber := ByteNumber + 1;
    od;
    writeString("    tile\ncorp;\n");
corp;

proc err(*char message; uint was, desired)void:

    if not LocalError then
	LocalError := true;
	writeString(":\n");
    fi;
    writeString(message);
    writeString(" - was ");
    writeCount(was);
    writeString(", must be ");
    writeCount(desired);
    writeString("\n");
corp;

proc readCat(*IFFContext_t context; IFFID_t subId)bool:

    IFFReadCatTail(context)
corp;

proc readList(*IFFContext_t context; IFFID_t subId)bool:
    ILBMPrivate_t ip;
    bool res;

    ip := pretend(context*.iff_client*.iff_userStructure, *ILBMPrivate_t)*;
    res := IFFReadListTail(context);
    pretend(context*.iff_client*.iff_userStructure, *ILBMPrivate_t)* := ip;
    res
corp;

proc readForm(*IFFContext_t context; IFFID_t subId)bool:
    ILBMPrivate_t ip;
    bool res;

    ip := pretend(context*.iff_client*.iff_userStructure, *ILBMPrivate_t)*;
    res := IFFReadFormTail(context);
    pretend(context*.iff_client*.iff_userStructure, *ILBMPrivate_t)* := ip;
    res
corp;

proc readProp(*IFFContext_t context; IFFID_t subId)bool:

    IFFReadPropTail(context)
corp;

proc readOther(*IFFContext_t context; IFFID_t id)bool:
    BitMap_t bm;
    *ILBMPrivate_t ip;
    uint depth, w, h, i;
    bool ok;

    ip := pretend(context*.iff_client*.iff_userStructure, *ILBMPrivate_t);
    case id.iff_ulong
    incase ID_BBMHD:
	if IFFReadBBMHD(context, &ip*.ip_bmh) then
	    if ip*.ip_bmh.bmh_nPlanes > DEPTH then
		err("too many bit planes", ip*.ip_bmh.bmh_nPlanes, DEPTH);
	    elif ip*.ip_bmh.bmh_nPlanes < DEPTH then
		err("too few bit planes", ip*.ip_bmh.bmh_nPlanes, DEPTH);
	    fi;
	    not LocalError
	else
	    false
	fi
    incase ID_BBODY:
	depth := ip*.ip_bmh.bmh_nPlanes;
	w := ip*.ip_bmh.bmh_w;
	h := ip*.ip_bmh.bmh_h;
	InitBitMap(&bm, depth, w, h);
	ok := true;
	for i from 0 upto depth - 1 do
	    bm.bm_Planes[i] := AllocRaster(w, h);
	    if bm.bm_Planes[i] = nil then
		ok := false;
	    fi;
	od;
	if ok then
	    ok := IFFReadBBODY(context, &bm, nil, &ip*.ip_bmh);
	    if ok then
		putImage(&bm, ip);
	    fi;
	fi;
	for i from depth - 1 downto 0 do
	    if bm.bm_Planes[i] ~= nil then
		FreeRaster(bm.bm_Planes[i], w, h);
	    fi;
	od;
	ok
    default:
	IFFSkipChunk(context)
    esac
corp;

proc printFailure(IFFError_t errorCode)void:

    writeString("Failed: ");
    writeString(
	case errorCode
	incase iff_ok:
	    "no error"
	incase iff_noFile:
	    "can't open file"
	incase iff_notIFF:
	    "not IFF file"
	incase iff_notEnough:
	    "file or chunk too short"
	incase iff_tooMuch:
	    "file or chunk too long"
	incase iff_badIFF:
	    "misformed IFF file"
	incase iff_illegalIFF:
	    "illegal structure for IFF file"
	default:
	    "unknown error"
	esac);
    writeString("\n");
corp;

proc usage()void:

    writeString("Use is: ILBMtoTile iff-file\n");
corp;

proc main()void:
    extern
	_d_pars_initialize()void,
	GetPar()*char;
    ILBMPrivate_t ip;
    IFFClient_t client;
    uint i;

    if OpenDosLibrary(0) ~= nil then
	if OpenExecLibrary(0) ~= nil then
	    if OpenGraphicsLibrary(0) ~= nil then
		StdOut := Output();
		_d_pars_initialize();
		client.iff_readForm := readForm;
		client.iff_readList := readList;
		client.iff_readCat := readCat;
		client.iff_readProp := readProp;
		client.iff_readOther := readOther;
		pretend(client.iff_userStructure, *ILBMPrivate_t) := &ip;
		InputName := GetPar();
		if InputName ~= nil then
		    ip.ip_bmh.bmh_w := 32;
		    ip.ip_bmh.bmh_h := 20;
		    ip.ip_bmh.bmh_nPlanes := 5;
		    ip.ip_bmh.bmh_masking := bmsk_none;
		    ip.ip_bmh.bmh_compression := bcmp_byteRun1;
		    LocalError := false;
		    if not IFFReadFile(InputName, &client) and not LocalError
		    then
			printFailure(client.iff_error);
		    fi;
		else
		    usage();
		fi;
		CloseGraphicsLibrary();
	    fi;
	    CloseExecLibrary();
	fi;
	CloseDosLibrary();
    fi;
corp;
