#drinc:libraries/dos.g
#iff.g

/*
 * IFFRead - attempt to read some bytes from an IFF context.
 */

proc IFFRead(register *IFFContext_t context; arbptr buffer;
	     register ulong length)bool:
    register *IFFClient_t client;
    register *byte p1, p2 @ context;
    register ulong actual;
    register bool ok;

    client := context*.iff_client;
    if length > context*.iff_totalRemaining then
	client*.iff_error := iff_notEnough;
	false
    else
	context*.iff_totalRemaining := context*.iff_totalRemaining - length;
	context*.iff_totalRead := context*.iff_totalRead + length;
	ok := true;
	if client*.iff_remaining ~= 0 then
	    if client*.iff_remaining > length then
		actual := length;
		client*.iff_remaining := client*.iff_remaining - length;
	    else
		actual := client*.iff_remaining;
		client*.iff_remaining := 0;
	    fi;
	    p1 := &client*.iff_data[client*.iff_position];
	    p2 := buffer;
	    client*.iff_position := client*.iff_position + actual;
	    length := length - actual;
	    buffer := pretend(buffer, *byte) + actual;
	    while actual ~= 0 do
		actual := actual - 1;
		p2* := p1*;
		p2 := p2 + 1;
		p1 := p1 + 1;
	    od;
	fi;
	if length ~= 0 then
	    if length >= IFF_BUFFER_SIZE then
		if Read(client*.iff_file, buffer, length) ~= length then
		    client*.iff_error := iff_notEnough;
		    ok := false;
		fi;
	    else
		client*.iff_position := 0;
		client*.iff_remaining :=
			Read(client*.iff_file, &client*.iff_data[0],
			     IFF_BUFFER_SIZE);
		if client*.iff_remaining < length then
		    client*.iff_error := iff_notEnough;
		    ok := false;
		else
		    p1 := &client*.iff_data[0];
		    p2 := buffer;
		    client*.iff_position := length;
		    client*.iff_remaining := client*.iff_remaining - length;
		    while length ~= 0 do
			length := length - 1;
			p2* := p1*;
			p2 := p2 + 1;
			p1 := p1 + 1;
		    od;
		fi;
	    fi;
	fi;
	ok
    fi
corp;

/*
 * IFFSkip - skip some bytes.
 */

proc IFFSkip(register *IFFContext_t context; register ulong amount)bool:
    register *IFFClient_t client;
    bool ok;

    if amount > context*.iff_totalRemaining then
	false
    else
	context*.iff_totalRemaining := context*.iff_totalRemaining - amount;
	context*.iff_totalRead := context*.iff_totalRead + amount;
	ok := true;
	client := context*.iff_client;
	if amount <= client*.iff_remaining then
	    client*.iff_remaining := client*.iff_remaining - amount;
	    client*.iff_position := client*.iff_position + amount;
	else
	    if Seek(client*.iff_file, amount - client*.iff_remaining,
		    OFFSET_CURRENT) = -1
	    then
		client*.iff_error := iff_notEnough;
		ok := false;
	    fi;
	    client*.iff_remaining := 0;
	    client*.iff_position := 0;
	fi;
	ok
    fi
corp;

/*
 * IFFSkipChunk - skip the current chunk.
 */

proc IFFSkipChunk(*IFFContext_t context)bool:

    IFFSkip(context, context*.iff_totalRemaining)
corp;

/*
 * IFFReadFile - open and read an IFF file.
 */

proc IFFReadFile(*char fileName; register *IFFClient_t client)bool:
    IFFContext_t context;
    IFFHeader2_t header;
    register bool ok;

    client*.iff_error := iff_ok;
    client*.iff_file := Open(fileName, MODE_READONLY);
    if client*.iff_file = 0 then
	client*.iff_error := iff_noFile;
	false
    else
	context.iff_client := client;
	context.iff_totalRemaining := sizeof(IFFHeader2_t);
	context.iff_totalRead := 0;
	client*.iff_position := 0;
	client*.iff_remaining := 0;
	ok := true;
	if IFFRead(&context, &header, sizeof(IFFHeader2_t)) then
	    context.iff_parent := nil;
	    /* allow for junk at the end of the file */
	    context.iff_totalRemaining :=
		    header.iff_header.iff_chunkLength - sizeof(IFFID_t);
	    case header.iff_header.iff_id.iff_ulong
	    incase ID_FORM:
		ok := client*.iff_readForm(&context, header.iff_subId);
	    incase ID_LIST:
		ok := client*.iff_readList(&context, header.iff_subId);
	    incase ID_CAT:
		ok := client*.iff_readCat(&context, header.iff_subId);
	    default:
		ok := false;
		client*.iff_error := iff_notIFF;
	    esac;
	    if ok then
		if context.iff_totalRemaining ~= 0 then
		    client*.iff_error := iff_tooMuch;
		    ok := false;
		fi;
	    fi;
	else
	    ok := false;
	fi;
	Close(client*.iff_file);
	ok
    fi
corp;

/*
 * IFFReadGroupClose - end of a recursive group for reading.
 */

proc IFFReadGroupClose(register *IFFContext_t context)bool:
    register *IFFContext_t parent;

    if context*.iff_totalRemaining = 0 then
	parent := context*.iff_parent;
	if parent ~= nil then
	    parent*.iff_totalRemaining :=
		parent*.iff_totalRemaining - context*.iff_totalRead;
	    parent*.iff_totalRead :=
		parent*.iff_totalRead + context*.iff_totalRead;
	fi;
	true
    else
	if context*.iff_client*.iff_error = iff_ok then
	    context*.iff_client*.iff_error := iff_tooMuch;
	fi;
	false
    fi
corp;

/*
 * IFFReadCatTail - read the remainder of an IFF CAT
 */

proc IFFReadCatTail(register *IFFContext_t parent)bool:
    IFFHeader2_t header;
    IFFContext_t child;
    register *IFFClient_t client;
    register bool ok;

    ok := true;
    client := parent*.iff_client;
    child.iff_parent := parent;
    child.iff_client := client;
    while parent*.iff_totalRemaining ~= 0 and ok do
	if IFFRead(parent, &header, sizeof(IFFHeader2_t)) then
	    if header.iff_header.iff_chunkLength - sizeof(IFFID_t) >
		parent*.iff_totalRemaining
	    then
		client*.iff_error := iff_badIFF;
		ok := false;
	    else
		child.iff_totalRemaining :=
			header.iff_header.iff_chunkLength - sizeof(IFFID_t);
		child.iff_totalRead := sizeof(IFFID_t);
		case header.iff_header.iff_id.iff_ulong
		incase ID_FORM:
		    ok := client*.iff_readForm(&child, header.iff_subId);
		incase ID_LIST:
		    ok := client*.iff_readList(&child, header.iff_subId);
		incase ID_CAT:
		    ok := client*.iff_readCat(&child, header.iff_subId);
		default:
		    ok := false;
		    client*.iff_error := iff_illegalIFF;
		esac;
	    fi;
	else
	    ok := false;
	fi;
    od;
    IFFReadGroupClose(parent) and ok
corp;

/*
 * IFFReadListTail - read the remainder of an IFF LIST
 */

proc IFFReadListTail(register *IFFContext_t parent)bool:
    IFFHeader2_t header;
    IFFContext_t child;
    register *IFFClient_t client;
    register bool ok, hadNonProp;

    ok := true;
    hadNonProp := false;
    client := parent*.iff_client;
    child.iff_parent := parent;
    child.iff_client := client;
    while parent*.iff_totalRemaining ~= 0 and ok do
	if IFFRead(parent, &header, sizeof(IFFHeader2_t)) then
	    if header.iff_header.iff_chunkLength - sizeof(IFFID_t) >
		parent*.iff_totalRemaining
	    then
		client*.iff_error := iff_badIFF;
		ok := false;;
	    else
		child.iff_totalRemaining :=
			header.iff_header.iff_chunkLength - sizeof(IFFID_t);
		child.iff_totalRead := sizeof(IFFID_t);
		case header.iff_header.iff_id.iff_ulong
		incase ID_PROP:
		    ok :=
			if hadNonProp then
			    false
			else
			    hadNonProp := true;
			    client*.iff_readProp(&child, header.iff_subId)
			fi;
		incase ID_FORM:
		    ok := client*.iff_readForm(&child, header.iff_subId);
		incase ID_LIST:
		    ok := client*.iff_readList(&child, header.iff_subId);
		incase ID_CAT:
		    ok := client*.iff_readCat(&child, header.iff_subId);
		default:
		    client*.iff_error := iff_illegalIFF;
		    ok := false;
		esac;
	    fi;
	else
	    ok := false;
	fi;
    od;
    IFFReadGroupClose(parent) and ok
corp;

/*
 * IFFReadFormTail - read the remainder of an IFF FORM
 */

proc IFFReadFormTail(register *IFFContext_t parent)bool:
    IFFHeader_t header;
    IFFContext_t child;
    IFFID_t subId;
    register *IFFClient_t client;
    register bool ok;

    ok := true;
    client := parent*.iff_client;
    child.iff_parent := parent;
    child.iff_client := client;
    while parent*.iff_totalRemaining ~= 0 and ok do
	if IFFRead(parent, &header, sizeof(IFFHeader_t)) then
	    if header.iff_chunkLength > parent*.iff_totalRemaining then
		client*.iff_error := iff_badIFF;
		ok := false;
	    else
		child.iff_totalRemaining := header.iff_chunkLength;
		child.iff_totalRead := 0;
		case header.iff_id.iff_ulong
		incase ID_FORM:
		    if not IFFRead(&child, &subId, sizeof(IFFID_t)) or
			not client*.iff_readForm(&child, subId)
		    then
			ok := false;
		    fi;
		incase ID_LIST:
		    if not IFFRead(&child, &subId, sizeof(IFFID_t)) or
			not client*.iff_readList(&child, subId)
		    then
			ok := false;
		    fi;
		incase ID_CAT:
		    if not IFFRead(&child, &subId, sizeof(IFFID_t)) or
			not client*.iff_readCat(&child, subId)
		    then
			ok := false;
		    fi;
		incase ID_PROP:
		    if not IFFRead(&child, &subId, sizeof(IFFID_t)) or
			not client*.iff_readProp(&child, subId)
		    then
			ok := false;
		    fi;
		default:
		    ok := client*.iff_readOther(&child, header.iff_id);
		    if header.iff_chunkLength & 1 ~= 0 then
			child.iff_totalRemaining :=
			    child.iff_totalRemaining + 1;
			if not IFFSkip(&child, 1) then
			    ok := false;
			fi;
		    fi;
		    if not IFFReadGroupClose(&child) then
			ok := false;
		    fi;
		esac;
	    fi;
	else
	    ok := false;
	fi;
    od;
    IFFReadGroupClose(parent) and ok
corp;

/*
 * IFFReadPropTail - read the remainder of an IFF PROP
 */

proc IFFReadPropTail(register *IFFContext_t parent)bool:
    IFFHeader_t header;
    IFFContext_t child;
    register *IFFClient_t client;
    register bool ok;

    ok := true;
    client := parent*.iff_client;
    child.iff_parent := parent;
    child.iff_client := client;
    while parent*.iff_totalRemaining ~= 0 and ok do
	if IFFRead(parent, &header, sizeof(IFFHeader_t)) then
	    if header.iff_chunkLength > parent*.iff_totalRemaining then
		client*.iff_error := iff_badIFF;
		ok := false;;
	    else
		child.iff_totalRemaining := header.iff_chunkLength;
		child.iff_totalRead := 0;
		ok := client*.iff_readOther(&child, header.iff_id);
		if header.iff_chunkLength & 1 ~= 0 then
		    if not IFFSkip(&child, 1) then
			ok := false;
		    fi;
		fi;
		if not IFFReadGroupClose(&child) then
		    ok := false;
		fi;
	    fi;
	else
	    ok := false;
	fi;
    od;
    IFFReadGroupClose(parent) and ok
corp;
