#drinc:libraries/dos.g
#drinc:libraries/dosextens.g
#drinc:exec/miscellaneous.g
#drinc:exec/errors.g
#drinc:exec/nodes.g
#drinc:exec/ports.g
#drinc:exec/io.g
#drinc:exec/memory.g
#drinc:exec/tasks.g
#drinc:devices/audio.g
#drinc:workbench/startup.g
#iff.g
#iff8SVX.g

type
    Private_t = struct {
	Voice8Header_t p_v8h;
	EGPoint_t p_attack, p_release;
    };

Handle_t StdOut, StdIn;
bool WorkBench, StdOpen, StdTried, HadError, LocalError, DescribeOnly,
     CurrentIs8SVX, Had8SVX;

proc writeString(*char str)void:
    register *char p;

    if WorkBench then
	if not StdOpen and not StdTried then
	    StdIn := Open("RAW:5/15/320/50/play8SVX", MODE_NEWFILE);
	    StdOut := StdIn;
	    if StdOut ~= 0 then
		StdOpen := true;
	    else
		StdTried := true;
	    fi;
	fi;
    fi;
    if StdOpen then
	p := str;
	while p* ~= '\e' do
	    p := p + 1;
	od;
	ignore Write(StdOut, str, p - str);
    fi;
corp;

proc writeCount(register ulong count)void:
    [11] char buffer;
    register *char p;

    p := &buffer[10];
    p* := '\e';
    while
	p := p - 1;
	p* := count % 10 + '0';
	count := count / 10;
	count ~= 0
    do
    od;
    writeString(p);
corp;

proc writeError(*char message)void:

    writeString(message);
    HadError := true;
    LocalError := true;
corp;

proc describeSample(register *Private_t p)void:

    writeString("oneShotHiSamples = ");
    writeCount(p*.p_v8h.v8h_oneShotHiSamples);
    writeString("\nrepeatHiSamples = ");
    writeCount(p*.p_v8h.v8h_repeatHiSamples);
    writeString("\n    (Total ");
    writeCount(p*.p_v8h.v8h_oneShotHiSamples + p*.p_v8h.v8h_repeatHiSamples);
    writeString(")\nsamplesPerHiCycle = ");
    writeCount(p*.p_v8h.v8h_samplesPerHiCycle);
    writeString("\nsamplesPerSecond = ");
    writeCount(p*.p_v8h.v8h_samplesPerSecond);
    writeString("\nctOctave = ");
    writeCount(p*.p_v8h.v8h_ctOctave);
    writeString("\ncompression = ");
    writeCount(p*.p_v8h.v8h_compression - scmp_none);
    writeString("\nvolume = ");
    writeCount(p*.p_v8h.v8h_volume);
    writeString("\nattack duration = ");
    writeCount(p*.p_attack.egp_duration);
    writeString("\nattack destination = ");
    writeCount(p*.p_attack.egp_dest);
    writeString("\nrelease duration = ");
    writeCount(p*.p_release.egp_duration);
    writeString("\nrelease destination = ");
    writeCount(p*.p_release.egp_dest);
    writeString("\n");
corp;

proc writeAudioError(*char message; long errorCode)void:

    writeError(message);
    writeString(": ");
    if errorCode = ADIOERR_NOALLOCATION then
	writeString("no audio channel allocated\n");
    elif errorCode = ADIOERR_ALLOCFAILED then
	writeString("allocation of sound channel failed\n");
    elif errorCode = ADIOERR_CHANNELSTOLEN then
	writeString("audio channel stolen\n");
    elif errorCode = IOERR_OPENFAIL then
	writeString("cannot open audio device\n");
    else
	writeString("other error ");
	writeCount(errorCode);
	writeString(" on audio device\n");
    fi;
corp;

proc runSample(*Private_t p; *short buffer; register ulong bufferLen)void:
    ulong MAX_AUDIO = 1 << 17;
    [4] short ALLOCATION = (1, 2, 4, 8);
    register *IOAudio_t ioa1, ioa2, ioaTemp;
    *MsgPort_t port;
    register ulong thisLen;
    long errorCode;

    ioa1 := AllocMem(sizeof(IOAudio_t), MEMF_CLEAR | MEMF_PUBLIC);
    if ioa1 = nil then
	writeError("Cannot allocate audio packet\n");
    else
	ioa2 := AllocMem(sizeof(IOAudio_t), MEMF_CLEAR | MEMF_PUBLIC);
	if ioa2 = nil then
	    writeError("Cannot allocate audio packet\n");
	else
	    port := CreatePort(nil, 0);
	    if port = nil then
		writeError("Cannot create audio reply port\n");
	    else
		ioa1*.ioa_Request.io_Message.mn_ReplyPort := port;
		ioa1*.ioa_Data := &ALLOCATION[0];
		ioa1*.ioa_Length := 4;
		ioa1*.ioa_Request.io_Message.mn_Node.ln_Pri := 0;
		errorCode := OpenDevice(AUDIONAME, 0, &ioa1*.ioa_Request, 0);
		if errorCode ~= 0 then
		    writeAudioError("OpenDevice", errorCode);
		else
		    ioa1*.ioa_Request.io_Command := CMD_WRITE;
		    ioa1*.ioa_Request.io_Flags := ADIOF_PERVOL;
		    ioa1*.ioa_Period := 1000000000 /
			    make(p*.p_v8h.v8h_samplesPerSecond, ulong) / 279;
		    ioa1*.ioa_Volume := p*.p_v8h.v8h_volume >> 10;
		    ioa1*.ioa_Cycles := 1;
		    ioa2* := ioa1*;
		    bufferLen := bufferLen & (-2);
		    thisLen :=
			if bufferLen > MAX_AUDIO then
			    MAX_AUDIO
			else
			    bufferLen
			fi;
		    ioa1*.ioa_Data := buffer;
		    ioa1*.ioa_Length := thisLen;
		    BeginIO(&ioa1*.ioa_Request);
		    bufferLen := bufferLen - thisLen;
		    buffer := buffer + thisLen;
		    while bufferLen ~= 0 do
			thisLen :=
			    if bufferLen > MAX_AUDIO then
				MAX_AUDIO
			    else
				bufferLen
			    fi;
			ioa2*.ioa_Data := buffer;
			ioa2*.ioa_Length := thisLen;
			BeginIO(&ioa2*.ioa_Request);
			bufferLen := bufferLen - thisLen;
			buffer := buffer + thisLen;
			errorCode := WaitIO(&ioa1*.ioa_Request);
			if errorCode ~= 0 then
			    writeAudioError("WaitIO1", errorCode);
			    bufferLen := 0;
			fi;
			ioaTemp := ioa1;
			ioa1 := ioa2;
			ioa2 := ioaTemp;
		    od;
		    errorCode := WaitIO(&ioa1*.ioa_Request);
		    if errorCode ~= 0 then
			writeAudioError("WaitIO2", errorCode);
		    fi;
		    CloseDevice(&ioa1*.ioa_Request);
		fi;
		DeletePort(port);
	    fi;
	    FreeMem(ioa2, sizeof(IOAudio_t));
	fi;
	FreeMem(ioa1, sizeof(IOAudio_t));
    fi;
corp;

proc readCat(*IFFContext_t context; IFFID_t subId)bool:

    IFFReadCatTail(context)
corp;

proc readList(*IFFContext_t context; IFFID_t subId)bool:
    Private_t p;
    bool res;

    p := pretend(context*.iff_client*.iff_userStructure, *Private_t)*;
    res := IFFReadListTail(context);
    pretend(context*.iff_client*.iff_userStructure, *Private_t)* := p;
    res
corp;

proc readForm(*IFFContext_t context; IFFID_t subId)bool:
    Private_t p;
    bool res;

    CurrentIs8SVX := subId.iff_ulong = ID_S8SVX;
    p := pretend(context*.iff_client*.iff_userStructure, *Private_t)*;
    res := IFFReadFormTail(context);
    pretend(context*.iff_client*.iff_userStructure, *Private_t)* := p;
    res
corp;

proc readProp(*IFFContext_t context; IFFID_t subId)bool:

    IFFReadPropTail(context)
corp;

proc readOther(register *IFFContext_t context; IFFID_t id)bool:
    register *Private_t p;
    *short buffer;
    ulong len;

    p := pretend(context*.iff_client*.iff_userStructure, *Private_t);
    case id.iff_ulong
    incase ID_SVHDR:
	IFFReadSVHDR(context, &p*.p_v8h)
    incase ID_SBODY:
	if CurrentIs8SVX then
	    Had8SVX := true;
	    len := p*.p_v8h.v8h_oneShotHiSamples +
		   p*.p_v8h.v8h_repeatHiSamples;
	    len := len * ((1 << p*.p_v8h.v8h_ctOctave) - 1);
	    buffer := AllocMem(len, MEMF_CHIP);
	    if buffer = nil then
		writeError("Can't allocate chip memory for buffer!\n");
		false
	    elif IFFReadSBODY(context, buffer, &len, &p*.p_v8h) then
		if DescribeOnly then
		    describeSample(p);
		else
		    runSample(p, buffer, len);
		fi;
		FreeMem(buffer, len);
		true
	    else
		false
	    fi
	else
	    IFFSkipChunk(context)
	fi
    default:
	IFFSkipChunk(context)
    esac
corp;

proc writeFailure(IFFError_t errorCode)void:

    writeError("Failed - ");
    writeString(
	case errorCode
	incase iff_ok:
	    "no error"
	incase iff_noFile:
	    "can't open file"
	incase iff_notIFF:
	    "not IFF file"
	incase iff_notEnough:
	    "file or chunk too short"
	incase iff_tooMuch:
	    "file or chunk too long"
	incase iff_badIFF:
	    "misformed IFF file"
	incase iff_illegalIFF:
	    "illegal structure for IFF file"
	default:
	    "unknown error"
	esac);
    writeString("\n");
corp;

proc playSample(*char name)void:
    Private_t p;
    register *char cp;
    IFFClient_t client;

    if not WorkBench then
	if name* = '"' then
	    cp := name;
	    cp := cp + sizeof(char);
	    name := cp;
	    while cp* ~= '"' and cp* ~= '\e' do
		cp := cp + sizeof(char);
	    od;
	    if cp* = '"' then
		cp* := '\e';
	    fi;
	fi;
	writeString(name);
	writeString(":");
	writeString("\n");
    fi;
    client.iff_readForm := readForm;
    client.iff_readList := readList;
    client.iff_readCat := readCat;
    client.iff_readProp := readProp;
    client.iff_readOther := readOther;
    pretend(client.iff_userStructure, *Private_t) := &p;
    p.p_v8h.v8h_oneShotHiSamples := 0;
    p.p_v8h.v8h_repeatHiSamples := 0;
    p.p_v8h.v8h_samplesPerHiCycle := 0;
    p.p_v8h.v8h_samplesPerSecond := 8363;
    p.p_v8h.v8h_ctOctave := 0;
    p.p_v8h.v8h_compression := scmp_none;
    p.p_v8h.v8h_volume := UNITY;
    p.p_attack.egp_duration := 0;
    p.p_attack.egp_dest := UNITY;
    p.p_release.egp_duration := 0;
    p.p_release.egp_dest := UNITY;
    LocalError := false;
    CurrentIs8SVX := false;
    Had8SVX := false;
    if IFFReadFile(name, &client) then
	if Had8SVX then
	    if not WorkBench then
		writeString("Success\n");
	    fi;
	else
	    HadError := true;
	    if WorkBench then
		writeString(name);
		writeString(":");
		writeString("\n");
	    fi;
	    writeString("No FORM 8SVX found.\n");
	fi;
    elif not LocalError then
	writeFailure(client.iff_error);
    fi;
corp;

proc main()void:
    extern
	_d_pars_initialize()void,
	GetPar()*char;
    *Process_t thisProcess;
    *Message_t mess;
    register *WBStartup_t sm;
    register *WBArg_t wa;
    register *char par @ sm;
    Lock_t oldDir;
    register uint i;
    byte b;

    if OpenDosLibrary(0) ~= nil then
	if OpenExecLibrary(0) ~= nil then
	    HadError := false;
	    thisProcess := pretend(FindTask(nil), *Process_t);
	    if thisProcess*.pr_CLI = 0 then
		/* running from WorkBench */
		WorkBench := true;
		StdOpen := false;
		StdTried := false;
		mess := WaitPort(&thisProcess*.pr_MsgPort);
		mess := GetMsg(&thisProcess*.pr_MsgPort);
		sm := pretend(mess, *WBStartup_t);
		wa := sm*.sm_ArgList;
		if wa ~= nil and sm*.sm_NumArgs > 1 then
		    for i from 1 upto sm*.sm_NumArgs - 1 do
			wa := wa + sizeof(WBArg_t);
			if wa*.wa_Lock ~= 0 then
			    oldDir := CurrentDir(wa*.wa_Lock);
			    playSample(wa*.wa_Name);
			    ignore CurrentDir(oldDir);
			fi;
		    od;
		else
		    writeError("No sounds to play!\n");
		fi;
		if HadError then
		    while Read(StdIn, &b, 1) = 1 and b ~= '\r' - '\e' do
		    od;
		fi;
		if StdOpen then
		    Close(StdIn);
		fi;
		Forbid();
		ReplyMsg(&sm*.sm_Message);
	    else
		/* running from CLI */
		WorkBench := false;
		StdOut := Output();
		StdIn := Input();
		StdOpen := true;
		_d_pars_initialize();
		DescribeOnly := false;
		while
		    par := GetPar();
		    par ~= nil
		do
		    if par* = '-' then
			while
			    par := par + sizeof(char);
			    par* ~= '\e'
			do
			    case par*
			    incase 'd':
			    incase 'D':
				DescribeOnly := true;
			    default:
				writeString("Unknown flag, only 'd' known\n");
			    esac;
			od;
		    else
			playSample(par);
		    fi;
		od;
	    fi;
	    CloseExecLibrary();
	fi;
	CloseDosLibrary();
    fi;
corp;
