#drinc:libraries/dos.g
#drinc:libraries/dosextens.g
#drinc:exec/miscellaneous.g
#drinc:exec/nodes.g
#drinc:exec/ports.g
#drinc:exec/io.g
#drinc:exec/interrupts.g
#drinc:exec/memory.g
#drinc:exec/tasks.g
#drinc:devices/timer.g
#drinc:hardware/intbits.g
#drinc:graphics/gfx.g
#drinc:graphics/view.g
#drinc:graphics/gfxbase.g
#drinc:graphics/rastport.g
#drinc:graphics/copper.g
#drinc:workbench/startup.g
#iff.g
#iffILBM.g

bool DRACO2 = true;

uint
    COLOR_TABLE_SIZE = 32,
    COLOR_CYCLE_SIZE = 4;

type
    Timer_t = struct {
	*MsgPort_t t_timerPort;
	*timerequest_t t_timerRequest;
	*Interrupt_t t_timerInterrupt;
	bool t_deviceOpen;
    },

    Cycler_t = struct {
	[range(bool)] Timer_t cy_timer;
	CRange_t cy_cRange;
	bool cy_parity, cy_active;
    },
	
    ILBMPrivate_t = struct {
	BitMapHeader_t ip_bmh;
	[COLOR_TABLE_SIZE] uint ip_colorMap;
	Point2D_t ip_p2d;
	DestMerge_t ip_dm;
	SpritePrecedence_t ip_sp;
	ViewPortMode_t ip_vpm;
	[COLOR_CYCLE_SIZE] Cycler_t ip_cycle;
	ushort ip_colorCount, ip_cycleCount;
	bool ip_gotCAMG;
    };

Handle_t StdOut, StdIn;

*ILBMPrivate_t CurrentIP;

*GfxBase_t GfxBase;
ViewPort_t ViewPort;

[COLOR_TABLE_SIZE] uint CurrentMap;

bool StatsOnly, CyclingEnabled, HadError, CurrentIsILBM, HadILBM;

proc writeString(*char str)void:
    register *char p;

    p := str;
    while p* ~= '\e' do
	p := p + 1;
    od;
    ignore Write(StdOut, str, p - str);
corp;

proc writeNumber(register ulong n)void:
    [11] char buffer;
    register *char p;

    p := &buffer[10];
    p* := '\e';
    while
	p := p - sizeof(char);
	p* := n % 10 + '0';
	n := n / 10;
	n ~= 0
    do
    od;
    writeString(p);
corp;

proc timerEnable(register *Cycler_t cy)void:
    register ulong millis;
    register *timerequest_t tr;

    millis := 16384 * 1000 / 60 / cy*.cy_cRange.cr_rate;
    tr := cy*.cy_timer[cy*.cy_parity].t_timerRequest;
    tr*.tr_time.tv_secs := millis / 1000;
    tr*.tr_time.tv_micro := millis * 1000;
    tr*.tr_node.io_Command := TR_ADDREQUEST;
    SendIO(&tr*.tr_node);
    cy*.cy_active := true;
corp;

proc timerHandler()void:
    uint
	OP_MOVEL = 0x2000,
	M_ADIR = 0o1,
	M_DISP = 0o5,
	R_FP = 6,
	R_SP = 7,
	A1 = 1;
    *Cycler_t cy;
    register uint i;
    uint low, high, temp;

    if DRACO2 then
	code(
	    OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_ADIR << 3 | A1,
	    cy
	);
    else
	code(
	    OP_MOVEL | R_SP << 9 | M_DISP << 6 | M_ADIR << 3 | A1,
	    cy
	);
    fi;
    cy*.cy_active := false;
    if CyclingEnabled then
	low := cy*.cy_cRange.cr_low;
	high := cy*.cy_cRange.cr_high;
	temp := CurrentMap[high];
	for i from high downto low + 1 do
	    CurrentMap[i] := CurrentMap[i - 1];
	od;
	CurrentMap[low] := temp;
	LoadRGB4(&ViewPort, &CurrentMap[0], CurrentIP*.ip_colorCount);
	cy*.cy_parity := not cy*.cy_parity;
	timerEnable(cy);
    fi;
corp;

proc timerDelete(register *Timer_t t)void:

    if t*.t_timerInterrupt ~= nil then
	FreeMem(t*.t_timerInterrupt, sizeof(Interrupt_t));
    fi;
    if t*.t_deviceOpen then
	CloseDevice(&t*.t_timerRequest*.tr_node);
    fi;
    if t*.t_timerRequest ~= nil then
	DeleteExtIO(&t*.t_timerRequest*.tr_node, sizeof(timerequest_t));
    fi;
    if t*.t_timerPort ~= nil then
	DeletePort(t*.t_timerPort);
    fi;
corp;

proc cyclerDelete(register *Cycler_t cy)void:

    if cy*.cy_active then
	if AbortIO(&cy*.cy_timer[cy*.cy_parity].t_timerRequest*.tr_node) ~= 0
	then
	    writeString("AbortIO on timer failed\n");
	    HadError := true;
	fi;
	Delay(5);	/* allow it to be killed and go away */
    fi;
    timerDelete(&cy*.cy_timer[false]);
    timerDelete(&cy*.cy_timer[true]);
corp;

proc timerCreate(*Cycler_t cy; register *Timer_t t)void:
    register *Interrupt_t i;

    t*.t_timerRequest := nil;
    t*.t_timerInterrupt := nil;
    t*.t_deviceOpen := false;
    t*.t_timerPort := CreatePort(nil, 0);
    if t*.t_timerPort ~= nil then
	t*.t_timerRequest :=
	    pretend(CreateExtIO(t*.t_timerPort, sizeof(timerequest_t)),
		    *timerequest_t);
	if t*.t_timerRequest ~= nil then
	    if OpenDevice(TIMERNAME, UNIT_VBLANK,
			  &t*.t_timerRequest*.tr_node, 0) = 0 then
		t*.t_deviceOpen := true;
		i := AllocMem(sizeof(Interrupt_t), MEMF_PUBLIC);
		if i ~= nil then
		    t*.t_timerInterrupt := i;
		    i*.is_Node.ln_Succ := nil;
		    i*.is_Node.ln_Pred := nil;
		    i*.is_Node.ln_Type := NT_INTERRUPT;
		    i*.is_Node.ln_Pri := 0;
		    i*.is_Node.ln_Name := nil;
		    i*.is_Code := timerHandler;
		    i*.is_Data := cy;
		    t*.t_timerPort*.mp_Flags :=
			t*.t_timerPort*.mp_Flags & ~PF_ACTION | PA_SOFTINT;
		    t*.t_timerPort*.mp_SigTask :=
			pretend(t*.t_timerInterrupt, *Task_t);
		fi;
	    fi;
	fi;
    fi;
corp;

proc cyclerCreate(register *Cycler_t cy)void:

    cy*.cy_parity := true;
    cy*.cy_active := false;
    timerCreate(cy, &cy*.cy_timer[false]);
    timerCreate(cy, &cy*.cy_timer[true]);
corp;

proc displayPicture(*BitMap_t bm)void:
    *View_t oldView;
    View_t view;
    RasInfo_t rasInfo;
    register uint i;
    byte b;

    oldView := GfxBase*.gb_ActiView;

    InitView(&view);
    view.v_ViewPort := &ViewPort;
    rasInfo.ri_BitMap := bm;
    rasInfo.ri_RxOffset := 0;
    rasInfo.ri_RyOffset := 0;
    InitVPort(&ViewPort);
    ViewPort.vp_DWidth := CurrentIP*.ip_bmh.bmh_w;
    ViewPort.vp_DHeight := CurrentIP*.ip_bmh.bmh_h;
    ViewPort.vp_RasInfo := &rasInfo;
    if ViewPort.vp_DWidth < 4 * 8 then
	ViewPort.vp_DWidth := 4 * 8;
    fi;
    if CurrentIP*.ip_gotCAMG then
	ViewPort.vp_Modes := CurrentIP*.ip_vpm &
	    ~(GENLOCK_VIDEO | PFBA | GENLOCK_AUDIO | VP_HIDE | SPRITES);
    else
	ViewPort.vp_Modes := 0x0;
    fi;
    if ViewPort.vp_DWidth > 400 then
	ViewPort.vp_Modes := ViewPort.vp_Modes | HIRES;
    fi;
    if ViewPort.vp_DHeight > 260 then
	ViewPort.vp_Modes := ViewPort.vp_Modes | LACE;
	view.v_Modes := view.v_Modes | LACE;
    fi;
    MakeVPort(&view, &ViewPort);
    MrgCop(&view);
    LoadView(&view);
    WaitTOF();
    CurrentMap := CurrentIP*.ip_colorMap;
    LoadRGB4(&ViewPort, &CurrentIP*.ip_colorMap[0], CurrentIP*.ip_colorCount);
    for i from 1 upto CurrentIP*.ip_cycleCount do
	cyclerCreate(&CurrentIP*.ip_cycle[i - 1]);
    od;
    CyclingEnabled := true;
    for i from 1 upto CurrentIP*.ip_cycleCount do
	timerEnable(&CurrentIP*.ip_cycle[i - 1]);
    od;

    while Read(StdIn, &b, 1) = 1 and b ~= '\r' - '\e' do
	if b = '\t' - '\e' then
	    CyclingEnabled := not CyclingEnabled;
	    if CyclingEnabled then
		for i from 1 upto CurrentIP*.ip_cycleCount do
		    if not CurrentIP*.ip_cycle[i - 1].cy_active then
			timerEnable(&CurrentIP*.ip_cycle[i - 1]);
		    fi;
		od;
	    else
		CurrentMap := CurrentIP*.ip_colorMap;
		LoadRGB4(&ViewPort, &CurrentIP*.ip_colorMap[0],
			 CurrentIP*.ip_colorCount);
	    fi;
	fi;
    od;

    CyclingEnabled := false;
    for i from 1 upto CurrentIP*.ip_cycleCount do
	cyclerDelete(&CurrentIP*.ip_cycle[i - 1]);
    od;
    LoadView(oldView);
    FreeVPortCopLists(&ViewPort);
    FreeCprList(view.v_LOFCprList);
corp;

proc readCat(*IFFContext_t context; IFFID_t subId)bool:

    IFFReadCatTail(context)
corp;

proc readList(*IFFContext_t context; IFFID_t subId)bool:
    ILBMPrivate_t ip;
    bool res;

    ip := pretend(context*.iff_client*.iff_userStructure, *ILBMPrivate_t)*;
    res := IFFReadListTail(context);
    pretend(context*.iff_client*.iff_userStructure, *ILBMPrivate_t)* := ip;
    res
corp;

proc readForm(*IFFContext_t context; IFFID_t subId)bool:
    ILBMPrivate_t ip;
    bool res;

    CurrentIsILBM := subId.iff_ulong = ID_BILBM;
    ip := pretend(context*.iff_client*.iff_userStructure, *ILBMPrivate_t)*;
    res := IFFReadFormTail(context);
    pretend(context*.iff_client*.iff_userStructure, *ILBMPrivate_t)* := ip;
    res
corp;

proc readProp(*IFFContext_t context; IFFID_t subId)bool:

    IFFReadPropTail(context)
corp;

proc writeColour(register uint c)void:
    [5] char buffer;
    register uint digit;

    buffer[4] := '\e';
    buffer[3] := ' ';
    digit := c & 0xf;
    buffer[2] := if digit >= 10 then digit - 10 + 'A' else digit + '0' fi;
    c := c >> 4;
    digit := c & 0xf;
    buffer[1] := if digit >= 10 then digit - 10 + 'A' else digit + '0' fi;
    c := c >> 4;
    digit := c & 0xf;
    buffer[0] := if digit >= 10 then digit - 10 + 'A' else digit + '0' fi;
    writeString(&buffer[0]);
corp;

proc readOther(register *IFFContext_t context; IFFID_t id)bool:
    BitMap_t bm;
    register *ILBMPrivate_t ip;
    register *CRange_t cr;
    register uint i;
    uint depth, w, h;
    bool ok, first;

writeString("readOther\n");
    ip := pretend(context*.iff_client*.iff_userStructure, *ILBMPrivate_t);
    case id.iff_ulong
    incase ID_BBMHD:
	if IFFReadBBMHD(context, &ip*.ip_bmh) then
writeString("other got BMHD OK\n");
	    if ip*.ip_bmh.bmh_nPlanes > 6 then
		ip*.ip_bmh.bmh_nPlanes := 6;
	    fi;
	    if StatsOnly then
		writeString("BitMap Header: w ");
		writeNumber(ip*.ip_bmh.bmh_w);
		writeString("  h ");
		writeNumber(ip*.ip_bmh.bmh_h);
		writeString("  nPlanes ");
		writeNumber(ip*.ip_bmh.bmh_nPlanes);
		writeString("\n  masking ");
		writeString(
		    case ip*.ip_bmh.bmh_masking
		    incase bmsk_none:
			"none"
		    incase bmsk_hasMask:
			"hasMask"
		    incase bmsk_hasTransParentColour:
			"hasTransparentColour"
		    incase bmsk_lasso:
			"lasso"
		    default:
			"???"
		    esac);
		writeString("  compression ");
		writeString(
		    case ip*.ip_bmh.bmh_compression
		    incase bcmp_none:
			"none"
		    incase bcmp_byteRun1:
			"byteRun1"
		    default:
			"???"
		    esac);
		writeString("  transparent color ");
		writeNumber(ip*.ip_bmh.bmh_transparentColor);
		writeString("\n  xAspect ");
		writeNumber(ip*.ip_bmh.bmh_xAspect);
		writeString("  yAspect ");
		writeNumber(ip*.ip_bmh.bmh_yAspect);
		writeString("  pageWidth ");
		writeNumber(ip*.ip_bmh.bmh_pageWidth);
		writeString("  pageHeight ");
		writeNumber(ip*.ip_bmh.bmh_pageHeight);
		writeString("\n");
	    fi;
	    true
	else
writeString("other: failed to read BMHD\n");
	    false
	fi
    incase ID_BCMAP:
	ip*.ip_colorCount := COLOR_TABLE_SIZE;
	if IFFReadBCMAP(context, &ip*.ip_colorMap[0], &ip*.ip_colorCount) then
writeString("other: got CMAP OK\n");
	    if StatsOnly then
		writeString("color map:\n  ");
		for i from 0 upto ip*.ip_colorCount - 1 do
		    writeColour(ip*.ip_colorMap[i]);
		    if i = 156 and ip*.ip_colorCount ~= 16 then
			writeString("\n  ");
		    fi;
		od;
		writeString("\n");
	    fi;
	    true
	else
writeString("other: failed to read CMAP\n");
	    false
	fi
    incase ID_BGRAB:
	IFFReadBGRAB(context, &ip*.ip_p2d)
    incase ID_BDEST:
	IFFReadBDEST(context, &ip*.ip_dm)
    incase ID_BSPRT:
	IFFReadBSPRT(context, &ip*.ip_sp)
    incase ID_BCAMG:
	if IFFReadBCAMG(context, &ip*.ip_vpm) then
writeString("other: got CAMG OK\n");
	    ip*.ip_gotCAMG := true;
	    if StatsOnly then
		writeString("CAMG: ");
		first := true;
		i := ip*.ip_vpm;
		if i & GENLOCK_VIDEO ~= 0 then
		    first := false;
		    writeString("genlock");
		fi;
		if i & LACE ~= 0 then
		    if not first then writeString(":") fi;
		    first := false;
		    writeString("lace");
		fi;
		if i & PFBA ~= 0 then
		    if not first then writeString(":") fi;
		    first := false;
		    writeString("pfba");
		fi;
		if i & EXTRA_HALFBRITE ~= 0 then
		    if not first then writeString(":") fi;
		    first := false;
		    writeString("halfbrite");
		fi;
		if i & GENLOCK_AUDIO ~= 0 then
		    if not first then writeString(":") fi;
		    first := false;
		    writeString("genAudio");
		fi;
		if i & DUALPF ~= 0 then
		    if not first then writeString(":") fi;
		    first := false;
		    writeString("dualpf");
		fi;
		if i & HAM ~= 0 then
		    if not first then writeString(":") fi;
		    first := false;
		    writeString("ham");
		fi;
		if i & VP_HIDE ~= 0 then
		    if not first then writeString(":") fi;
		    first := false;
		    writeString("vp-hide");
		fi;
		if i & SPRITES ~= 0 then
		    if not first then writeString(":") fi;
		    first := false;
		    writeString("sprites");
		fi;
		if i & HIRES ~= 0 then
		    if not first then writeString(":") fi;
		    first := false;
		    writeString("hires");
		fi;
		writeString("\n");
	    fi;
	    true
	else
writeString("other: failed to read CAMG\n");
	    false
	fi
    incase ID_BCRNG:
	if ip*.ip_cycleCount = COLOR_CYCLE_SIZE then
	    IFFSkipChunk(context)
	else
	    cr := &ip*.ip_cycle[ip*.ip_cycleCount].cy_cRange;
	    if IFFReadBCRNG(context, cr) then
		if cr*.cr_active ~= 0 and cr*.cr_low ~= cr*.cr_high then
		    ip*.ip_cycleCount := ip*.ip_cycleCount + 1;
		    if cr*.cr_low > cr*.cr_high then
			writeString("cr_low > cr_high\n");
			HadError := true;
			i := cr*.cr_low;
			cr*.cr_low := cr*.cr_high;
			cr*.cr_high := i;
		    fi;
		fi;
		true
	    else
		false
	    fi
	fi
    incase ID_BBODY:
	if CurrentIsILBM then
	    HadILBM := true;
	    depth := ip*.ip_bmh.bmh_nPlanes;
	    w := ip*.ip_bmh.bmh_w;
	    h := ip*.ip_bmh.bmh_h;
	    if StatsOnly then
writeString("other: BODY: StatsOnly\n");
		ok := IFFSkipChunk(context);
	    else
		InitBitMap(&bm, depth, w, h);
		ok := true;
		for i from depth - 1 downto 0 do
		    bm.bm_Planes[i] := AllocRaster(w, h);
		    if bm.bm_Planes[i] = nil then
			writeString("can't allocate chip RAM\n");
			HadError := true;
			ok := false;
		    fi;
		od;
		if ok then
		    ok := IFFReadBBODY(context, &bm, nil, &ip*.ip_bmh);
		    if ok then
writeString("other: got BODY ok\n");
			CurrentIP := ip;
			displayPicture(&bm);
else
writeString("other: failed to read BODY\n");
		    fi;
		else
		    ignore IFFSkipChunk(context);
		fi;
		for i from depth - 1 downto 0 do
		    if bm.bm_Planes[i] ~= nil then
			FreeRaster(bm.bm_Planes[i], w, h);
		    fi;
		od;
	    fi;
	    ok
	else
writeString("other: current not ILBM\n");
if
	    IFFSkipChunk(context)
then
    writeString("skipped BODY ok\n");
    true
else
    writeString("failed to skip body\n");
    false
fi
	fi
    default:
writeString("other skipping unknown\n");
	IFFSkipChunk(context)
    esac
corp;

proc writeFailure(IFFError_t errorCode)void:

    writeString("Failed - ");
    writeString(
	case errorCode
	incase iff_ok:
	    "no error"
	incase iff_noFile:
	    "can't open file"
	incase iff_notIFF:
	    "not IFF file"
	incase iff_notEnough:
	    "file or chunk too short"
	incase iff_tooMuch:
	    "file or chunk too long"
	incase iff_badIFF:
	    "misformed IFF file"
	incase iff_illegalIFF:
	    "illegal structure for IFF file"
	default:
	    "unknown error"
	esac);
    writeString("\n");
    HadError := true;
corp;

proc showPicture(*char name)void:
    ILBMPrivate_t ip;
    IFFClient_t client;
    register uint i;

    writeString(name);
    writeString(":");
    writeString("\n");
    client.iff_readForm := readForm;
    client.iff_readList := readList;
    client.iff_readCat := readCat;
    client.iff_readProp := readProp;
    client.iff_readOther := readOther;
    client.iff_userStructure:= &ip;
    ip.ip_bmh.bmh_w := 320;
    ip.ip_bmh.bmh_h := 200;
    ip.ip_bmh.bmh_nPlanes := 5;
    ip.ip_bmh.bmh_masking := bmsk_none;
    ip.ip_bmh.bmh_compression := bcmp_byteRun1;
    ip.ip_colorCount := COLOR_TABLE_SIZE;
    ip.ip_cycleCount := 0;
    for i from COLOR_CYCLE_SIZE - 1 downto 0 do
	ip.ip_cycle[i].cy_cRange.cr_active := 0;
    od;
    ip.ip_gotCAMG := false;
    CurrentIsILBM := false;
    HadILBM := false;
    if IFFReadFile(name, &client) then
	if HadILBM then
	    writeString("Success\n");
	else
	    writeString("No ILBM found\n");
	    HadError := true;
	fi;
    else
	writeFailure(client.iff_error);
    fi;
corp;

proc main()void:
    extern
	_d_pars_initialize()void,
	GetPar()*char;
    *Process_t thisProcess;
    *Message_t mess;
    register *WBStartup_t sm;
    register *WBArg_t wa;
    register *char par @ sm, p;
    Lock_t oldDir;
    register uint i;
    byte b;

    if OpenDosLibrary(0) ~= nil then
	if OpenExecLibrary(0) ~= nil then
	    GfxBase := OpenGraphicsLibrary(0);
	    if GfxBase ~= nil then
		HadError := false;
		StatsOnly := false;
		thisProcess := pretend(FindTask(nil), *Process_t);
		if thisProcess*.pr_CLI = 0 then
		    /* running from WorkBench */
		    mess := WaitPort(&thisProcess*.pr_MsgPort);
		    mess := GetMsg(&thisProcess*.pr_MsgPort);
		    sm := pretend(mess, *WBStartup_t);
		    StdIn := Open("RAW:5/15/320/50/showILBM", MODE_NEWFILE);
		    if StdIn ~= 0 then
			StdOut := StdIn;
			wa := sm*.sm_ArgList;
			if wa ~= nil and sm*.sm_NumArgs > 1 then
			    for i from 1 upto sm*.sm_NumArgs - 1 do
				wa := wa + sizeof(WBArg_t);
				if wa*.wa_Lock ~= 0 then
				    oldDir := CurrentDir(wa*.wa_Lock);
				    showPicture(wa*.wa_Name);
				    ignore CurrentDir(oldDir);
				fi;
			    od;
			else
			    writeString("No pictures to show!\n");
			    HadError := true;
			fi;
			if HadError then
			    while Read(StdIn, &b,1) = 1 and b ~= '\r' - '\e' do
			    od;
			fi;
			Close(StdIn);
		    fi;
		    Forbid();
		    ReplyMsg(&sm*.sm_Message);
		else
		    /* running from CLI */
		    StdOut := Output();
		    StdIn := Open("RAW:5/15/150/30/showILBM", MODE_NEWFILE);
		    if StdIn ~= 0 then
			_d_pars_initialize();
			while
			    par := GetPar();
			    par ~= nil
			do
			    if par* = '-' then
				StatsOnly := true;
				while par* ~= '\e' do
				    par := par + sizeof(char);
				od;
			    else
				if par* = '"' then
				    par := par + sizeof(char);
				    p := par;
				    while p* ~= '"' and p* ~= '\e' do
					p := p + sizeof(char);
				    od;
				    if p* = '"' then
					p* := '\e';
				    fi;
				fi;
				showPicture(par);
			    fi;
			od;
			Close(StdIn);
		    else
			writeString("Can't open RAW: window\n");
		    fi;
		fi;
		CloseGraphicsLibrary();
	    fi;
	    CloseExecLibrary();
	fi;
	CloseDosLibrary();
    fi;
corp;
