#drinc:libraries/dos.g
#drinc:exec/miscellaneous.g
#drinc:exec/nodes.g
#drinc:exec/ports.g
#drinc:exec/io.g
#drinc:exec/interrupts.g
#drinc:exec/memory.g
#drinc:devices/timer.g
#drinc:hardware/intbits.g
#drinc:graphics/gfx.g
#drinc:graphics/view.g
#drinc:graphics/gfxbase.g
#drinc:graphics/rastport.g
#drinc:graphics/copper.g
#drinc:util.g

bool DRACO2 = true;

uint
    IMAGE_BUFFER_SIZE = 10000,
    COLOR_TABLE_SIZE = 16,
    COLOR_CYCLE_SIZE = 4;

byte GOT_COLOR_MAP = 0x80;

type
    Timer_t = struct {
	*MsgPort_t t_timerPort;
	*timerequest_t t_timerRequest;
	*Interrupt_t t_timerInterrupt;
	bool t_deviceOpen;
    },

    Cycler_t = struct {
	[range(bool)] Timer_t cy_timer;
	uint cy_rate;
	ushort cy_low, cy_high;
	bool cy_parity, cy_active;
    };
	
[COLOR_CYCLE_SIZE] Cycler_t Cycler;
uint CycleCount;
[IMAGE_BUFFER_SIZE] byte Buffer;

*GfxBase_t GfxBase;
ViewPort_t ViewPort;
BitMap_t BitMap;
[COLOR_TABLE_SIZE] uint BaseMap, CurrentMap;
bool CyclingEnabled;

Handle_t StdOut, StdIn;

proc writeString(*char str)void:
    *char p;

    p := str;
    while p* ~= '\e' do
	p := p + 1;
    od;
    ignore Write(StdOut, str, p - str);
corp;

proc timerEnable(*Cycler_t cy)void:
    ulong millis;
    *timerequest_t tr;

    millis := 16384 * 1000 / 60 / cy*.cy_rate;
    tr := cy*.cy_timer[cy*.cy_parity].t_timerRequest;
    tr*.tr_time.tv_secs := millis / 1000;
    tr*.tr_time.tv_micro := millis * 1000;
    tr*.tr_node.io_Command := TR_ADDREQUEST;
    SendIO(&tr*.tr_node);
    cy*.cy_active := true;
corp;

proc timerHandler()void:
    uint
	OP_MOVEL = 0x2000,
	M_ADIR = 0o1,
	M_DISP = 0o5,
	R_FP = 6,
	R_SP = 7,
	A1 = 1;
    *Cycler_t cy;
    uint low, high, i, temp;

    if DRACO2 then
	code(
	    OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_ADIR << 3 | A1,
	    cy
	);
    else
	code(
	    OP_MOVEL | R_SP << 9 | M_DISP << 6 | M_ADIR << 3 | A1,
	    cy
	);
    fi;
    cy*.cy_active := false;
    if CyclingEnabled then
	low := cy*.cy_low;
	high := cy*.cy_high;
	temp := CurrentMap[high];
	for i from high downto low + 1 do
	    CurrentMap[i] := CurrentMap[i - 1];
	od;
	CurrentMap[low] := temp;
	LoadRGB4(&ViewPort, &CurrentMap[0], COLOR_TABLE_SIZE);
	cy*.cy_parity := not cy*.cy_parity;
	timerEnable(cy);
    fi;
corp;

proc timerDelete(*Timer_t t)void:

    if t*.t_timerInterrupt ~= nil then
	FreeMem(t*.t_timerInterrupt, sizeof(Interrupt_t));
    fi;
    if t*.t_deviceOpen then
	CloseDevice(&t*.t_timerRequest*.tr_node);
    fi;
    if t*.t_timerRequest ~= nil then
	DeleteExtIO(&t*.t_timerRequest*.tr_node, sizeof(timerequest_t));
    fi;
    if t*.t_timerPort ~= nil then
	DeletePort(t*.t_timerPort);
    fi;
corp;

proc cyclerDelete(*Cycler_t cy)void:

    if cy*.cy_active then
	if AbortIO(&cy*.cy_timer[cy*.cy_parity].t_timerRequest*.tr_node) ~= 0
	then
	    writeString("AbortIO on timer failed\n");
	fi;
	Delay(5);	/* allow it to be killed and go away */
    fi;
    timerDelete(&cy*.cy_timer[false]);
    timerDelete(&cy*.cy_timer[true]);
corp;

proc timerCreate(*Cycler_t cy; *Timer_t t)void:
    *Interrupt_t i;

    t*.t_timerRequest := nil;
    t*.t_timerInterrupt := nil;
    t*.t_deviceOpen := false;
    t*.t_timerPort := CreatePort(nil, 0);
    if t*.t_timerPort ~= nil then
	t*.t_timerRequest :=
	    pretend(CreateExtIO(t*.t_timerPort, sizeof(timerequest_t)),
		    *timerequest_t);
	if t*.t_timerRequest ~= nil then
	    if OpenDevice(TIMERNAME, UNIT_VBLANK,
			  &t*.t_timerRequest*.tr_node, 0) = 0 then
		t*.t_deviceOpen := true;
		i := AllocMem(sizeof(Interrupt_t), MEMF_PUBLIC);
		if i ~= nil then
		    t*.t_timerInterrupt := i;
		    i*.is_Node.ln_Succ := nil;
		    i*.is_Node.ln_Pred := nil;
		    i*.is_Node.ln_Type := NT_INTERRUPT;
		    i*.is_Node.ln_Pri := 0;
		    i*.is_Node.ln_Name := nil;
		    i*.is_Code := timerHandler;
		    i*.is_Data := pretend(cy, *byte);
		    t*.t_timerPort*.mp_Flags :=
			t*.t_timerPort*.mp_Flags & ~PF_ACTION | PA_SOFTINT;
		    t*.t_timerPort*.mp_SigTask :=
			pretend(t*.t_timerInterrupt, *Task_t);
		fi;
	    fi;
	fi;
    fi;
corp;

proc cyclerCreate(*Cycler_t cy)void:

    cy*.cy_parity := true;
    cy*.cy_active := false;
    timerCreate(cy, &cy*.cy_timer[false]);
    timerCreate(cy, &cy*.cy_timer[true]);
corp;

proc displayImage()void:
    *View_t oldView;
    View_t view;
    RasInfo_t rasInfo;
    uint i;
    byte b;

    oldView := GfxBase*.gb_ActiView;

    InitView(&view);
    view.v_ViewPort := &ViewPort;
    rasInfo.ri_BitMap := &BitMap;
    rasInfo.ri_RxOffset := 0;
    rasInfo.ri_RyOffset := 0;
    InitVPort(&ViewPort);
    ViewPort.vp_DWidth := 128;
    ViewPort.vp_DHeight := 128;
    ViewPort.vp_RasInfo := &rasInfo;
    ViewPort.vp_Modes := 0;
    MakeVPort(&view, &ViewPort);
    MrgCop(&view);
    LoadView(&view);
    WaitTOF();
    CurrentMap := BaseMap;
    LoadRGB4(&ViewPort, &BaseMap[0], COLOR_TABLE_SIZE);
    for i from 1 upto CycleCount do
	cyclerCreate(&Cycler[i - 1]);
    od;
    CyclingEnabled := true;
    for i from 1 upto CycleCount do
	timerEnable(&Cycler[i - 1]);
    od;

    while Read(StdIn, &b, 1) = 1 and b ~= '\r' - '\e' do
	if b = '\t' - '\e' then
	    CyclingEnabled := not CyclingEnabled;
	    if CyclingEnabled then
		for i from 1 upto CycleCount do
		    if not Cycler[i - 1].cy_active then
			timerEnable(&Cycler[i - 1]);
		    fi;
		od;
	    else
		CurrentMap := BaseMap;
		LoadRGB4(&ViewPort, &BaseMap[0], COLOR_TABLE_SIZE);
	    fi;
	fi;
    od;

    CyclingEnabled := false;
    for i from 1 upto CycleCount do
	cyclerDelete(&Cycler[i - 1]);
    od;
    LoadView(oldView);
    FreeVPortCopLists(&ViewPort);
    FreeCprList(view.v_LOFCprList);
corp;

proc readImage(*byte p)bool:
    *byte q;
    uint i, bytesLeft, j;
    int n2;
    short n;
    byte flags, b;
    bool ok;

    flags := p*;
    CycleCount := flags & ~GOT_COLOR_MAP;
    p := p + 1;
    if flags & GOT_COLOR_MAP ~= 0 then
	BlockCopy(&BaseMap[0], p, COLOR_TABLE_SIZE * sizeof(uint));
	p := p + COLOR_TABLE_SIZE * sizeof(uint);
    else
	/* put in default color map */
    fi;
    for i from 1 upto CycleCount do
	Cycler[i - 1].cy_rate := make(p*, uint) << 8 | (p + 1)*;
	Cycler[i - 1].cy_low := (p + 2)*;
	Cycler[i - 1].cy_high := (p + 3)*;
	p := p + 4;
    od;
    ok := true;
    InitBitMap(&BitMap, 4, 128, 128);
    for i from 0 upto 3 do
	BitMap.bm_Planes[i] := AllocRaster(128, 128);
	if BitMap.bm_Planes[i] = nil then
	    ok := false;
	fi;
    od;
    if ok then
	for i from 0 upto 3 do
	    bytesLeft := 128 * 128 / 8;
	    q := pretend(BitMap.bm_Planes[i], *byte);
	    while bytesLeft ~= 0 do
		n := p*;
		p := p + 1;
		n2 := n;
		if n2 < 0 then
		    n2 := - n2 + 1;
		    b := p*;
		    p := p + 1;
		    for j from 1 upto n2 do
			q* := b;
			q := q + 1;
		    od;
		else
		    n2 := n2 + 1;
		    BlockCopy(q, p, n2);
		    p := p + n2;
		    q := q + n2;
		fi;
		bytesLeft := bytesLeft - n2;
	    od;
	od;
	displayImage();
    fi;
    for i from 3 downto 0 do
	if BitMap.bm_Planes[i] ~= nil then
	    FreeRaster(BitMap.bm_Planes[i], 128, 128);
	fi;
    od;
    ok
corp;

proc main()void:
    extern
	_d_pars_initialize()void;
    register *char par, p;
    Handle_t fd;

    if OpenExecLibrary(0) ~= nil then
	if OpenDosLibrary(0) ~= nil then
	    GfxBase := OpenGraphicsLibrary(0);
	    if GfxBase ~= nil then
		StdOut := Output();
		StdIn := Open("RAW:0/0/150/30/showImage", MODE_NEWFILE);
		if StdIn ~= 0 then
		    _d_pars_initialize();
		    while
			par := GetPar();
			par ~= nil
		    do
			if par* = '"' then
			    par := par + sizeof(char);
			    p := par;
			    while p* ~= '"' and p* ~= '\e' do
				p := p + sizeof(char);
			    od;
			    if p* = '"' then
				p* := '\e';
			    fi;
			fi;
			writeString(par);
			writeString(":");
			writeString("\n");
			fd := Open(par, MODE_OLDFILE);
			if fd ~= 0 then
			    ignore Read(fd, &Buffer[0], IMAGE_BUFFER_SIZE);
			    Close(fd);
			    if readImage(&Buffer[0]) then
				writeString("Displayed ok\n");
			    else
				writeString("Can't display image\n");
			    fi;
			else
			    writeString("Can't open file\n");
			fi;
		    od;
		    Close(StdIn);
		else
		    writeString("Can't open RAW: window\n");
		fi;
		CloseGraphicsLibrary();
	    fi;
	    CloseDosLibrary();
	fi;
	CloseExecLibrary();
    fi;
corp;
